"""
Created on Thu Oct 16 12:00:00 2025

@author: Anna Grim
@email: anna.grim@alleninstitute.org

Code that provides a wrapper for loading data and calling the SkeletMetric
subclasses.

"""

from zipfile import ZipFile

import numpy as np
import os
import pandas as pd

from segmentation_skeleton_metrics.skeleton_metrics import (
    AddedCableLengthMetric,
    MergeCountMetric,
    MergeRateMetric,
    MergedEdgePercentMetric,
    OmitEdgePercentMetric,
    SplitEdgePercentMetric,
    SplitCountMetric,
    SplitRateMetric,
    EdgeAccuracyMetric,
    ERLMetric,
    NormalizedERLMetric,
)
from segmentation_skeleton_metrics.data_handling.graph_loading import (
    DataLoader,
    LabelHandler,
)
from segmentation_skeleton_metrics.utils import util


def evaluate(
    gt_path,
    segmentation,
    output_dir,
    anisotropy=(1.0, 1.0, 1.0),
    connections_path=None,
    fragments_path=None,
    results_filename="results",
    save_merges=False,
    save_fragments=False,
    use_anisotropy=False,
    valid_labels=None,
    verbose=True,
):
    """
    Loads data, calls an evaluator object that computes skeleton-based
    segmentation, and saves the results.

    Parameters
    ----------
    gt_path : str
        Path to ground truth SWC files, see swc_util.Reader for documentation
        These SWC files are assumed to be stored in voxel coordinates.
    segmentation : ImageReader
        Predicted segmentation.
    anisotropy : Tuple[float], optional
        Image to physical coordinates scaling factors to account for the
        anisotropy of the microscope. Default is (1.0, 1.0, 1.0).
    connections_path : str, optional
        Path to a txt file containing pairs of segment IDs that represents
        fragments that were merged. Default is None.
    fragments_path : str, optional
        Path to SWC files corresponding to "segmentation", see swc_util.Reader
        for documentation. Notes: (1) "anisotropy" is applied to these SWC
        files and (2) these SWC files are required for counting merges.
        Default is None.
    results_filename : str, optional
        Name of file that skeleton metric results are written to. Default is
        "results".
    save_merges : bool, optional
        Indication of whether to save merge sites and fragments with a merge
        mistake. Default is False.
    save_fragments : bool, optional
        Indication of whether to save fragments that intersect with each
        ground truth skeleton. Default is False.
    use_anisotropy : bool, optional
        Indication of whether to apply the anisotropy to the coordinates from
        the fragment SWC files. Default is False.
    valid_labels : Set[int], optional
        Labels that are allowed to be assigned. Default is None.
    verbose : bool, optional
        Indication of whether to display progress bars and printout results.
        Default is True.
    """
    # Load data
    label_handler = LabelHandler(connections_path, valid_labels)
    dataloader = DataLoader(
        label_handler,
        anisotropy=anisotropy,
        use_anisotropy=use_anisotropy,
        verbose=verbose,
    )
    gt_graphs = dataloader.load_groundtruth(gt_path, segmentation)
    fragment_graphs = dataloader.load_fragments(fragments_path, gt_graphs)

    # Run evaluation
    util.mkdir(output_dir)
    evaluator = Evaluator(output_dir, results_filename, verbose)
    evaluator(gt_graphs, fragment_graphs)

    # Optional saves
    if save_merges:
        evaluator.save_merge_results(gt_graphs, fragment_graphs, output_dir)

    if save_fragments and fragment_graphs:
        evaluator.save_fragments(gt_graphs, fragment_graphs)


# --- Evaluator ---
class Evaluator:
    """
    A class that evaluates neuron reconstruction quality by computing a set of
    skeleton-based metrics.

    Attributes
    ----------
    output_dir : str
        Directory where evaluation results will be saved.
    results_filename : str
        Filename (without extension) for the CSV report.
    verbose : bool
        Indication of whether to display progress bars and printout results.
    metrics : dict
        Core evaluation metrics mapping metric names to metric objects:
        - "# Splits": SplitCountMetric
        - "# Merges": MergeCountMetric
        - "% Split Edges": SplitEdgePercentMetric
        - "% Omit Edges": OmitEdgePercentMetric
        - "% Merged Edges": MergedEdgePercentMetric
        - "ERL": ERLMetric
    derived_metrics : dict
        Derived metrics computed from core metrics:
        - "Normalized ERL": NormalizedERLMetric
        - "Edge Accuracy": EdgeAccuracyMetric
        - "Split Rate": SplitRateMetric
        - "Merge Rate": MergeRateMetric
    """

    def __init__(self, output_dir, results_filename, verbose=True):
        """
        Instantiates an Evaluator object.

        Parameters
        ----------
        output_dir : str
            Directory where evaluation results will be saved.
        results_filename : str
            Filename (without extension) for the CSV report.
        verbose : bool, optional
            Indication of whether to display progress bars and printout
            results. Default is True.
        """
        # Instance attributes
        self.output_dir = output_dir
        self.results_filename = results_filename
        self.verbose = verbose

        # Set core metrics
        self.metrics = {
            "# Splits": SplitCountMetric(verbose=verbose),
            "# Merges": MergeCountMetric(verbose=verbose),
            "% Split Edges": SplitEdgePercentMetric(verbose=verbose),
            "% Omit Edges": OmitEdgePercentMetric(verbose=verbose),
            "% Merged Edges": MergedEdgePercentMetric(verbose=verbose),
            "ERL": ERLMetric(verbose=verbose),
        }

        # Set derived metrics
        self.derived_metrics = {
            "Normalized ERL": NormalizedERLMetric(verbose=verbose),
            "Edge Accuracy": EdgeAccuracyMetric(verbose=verbose),
            "Split Rate": SplitRateMetric(verbose=verbose),
            "Merge Rate": MergeRateMetric(verbose=verbose),
        }

    # --- Core Routines ---
    def __call__(self, gt_graphs, fragment_graphs=None):
        """
        Computes evaluation metrics for neuron reconstructions and saves a CSV
        report.

        Parameters
        ----------
        gt_graphs : Dict[str, LabeledGraph]
            Graphs to be evaluated.
        fragment_graphs : Dict[str, FragmentsGraph], optional
            Graphs built from skeletons obtained from a segmentation. This
            parameter is required to compute the metric "# Merges". Default
            is None.
        """
        # Printout step
        if self.verbose:
            print("\n(3) Compute Metrics")

        # Compute core metrics
        results = self.init_results(gt_graphs)
        for name, metric in self.metrics.items():
            if name == "# Merges" and fragment_graphs:
                results[name] = metric(gt_graphs, fragment_graphs)
            elif name != "# Merges":
                results.update(metric(gt_graphs))

        # Compute derived metrics
        for name, metric in self.derived_metrics.items():
            if name == "Merge Rate" and fragment_graphs:
                results[name] = metric(gt_graphs, results)
            elif name != "Merge Rate":
                results[name] = metric(gt_graphs, results)

        # Compute special metrics
        metric = AddedCableLengthMetric(verbose=self.verbose)
        metric(
            gt_graphs, fragment_graphs, self.metrics["# Merges"].merge_sites
        )

        # Save report
        path = f"{self.output_dir}/{self.results_filename}.csv"
        results.to_csv(path, index=True)
        self.report_summary(results)

    def init_results(self, gt_graphs):
        """
        Initializes a data frame that results from skeleton metrics will be
        stored in.

        Parameters
        ----------
        gt_graphs : Dict[str, LabeledGraph]
            Graphs to be evaluated.

        Returns
        -------
        results : pandas.DataFrame
            Data frame that results from skeleton metrics will be stored in.
        """
        # Create dataframe
        cols = (
            ["SWC Run Length"]
            + list(self.metrics.keys())
            + list(self.derived_metrics.keys())
        )
        index = list(gt_graphs.keys())
        index.sort()
        results = pd.DataFrame(np.nan, index=index, columns=cols)

        # Populate SWC Run Length column
        for key, graph in gt_graphs.items():
            results.loc[key, "SWC Run Length"] = graph.run_length
        return results

    def report_summary(self, results):
        """
        Generates and saves a summary report of evaluation results.

        Parameters
        ----------
        results : pandas.DataFrame
            DataFrame containing evaluation results for individual SWCs.
        """
        # Averaged results
        filename = f"{self.results_filename}-overview.txt"
        path = os.path.join(self.output_dir, filename)
        util.update_txt(path, "\nAverage Results...", self.verbose)
        for column in results.columns:
            if column != "SWC Run Length" and column != "SWC Name":
                avg = util.compute_weighted_avg(results, column)
                util.update_txt(path, f"  {column}: {avg:.4f}", self.verbose)

        # Total results
        n_splits = int(results["# Splits"].sum())
        util.update_txt(path, "\nTotal Results...", self.verbose)
        util.update_txt(path, f"  # Splits: {n_splits}", self.verbose)
        if "# Merges" in results.columns:
            n_merges = results["# Merges"].sum()
            util.update_txt(path, f"  # Merges: {n_merges}", self.verbose)

    # --- Writers ---
    def save_fragments(self, gt_graphs, fragment_graphs):
        """
        Saves ground-truth graphs and their intersecting fragment graphs to
        zipped SWC files.

        Parameters
        ----------
        gt_graphs : Dict[str, LabeledGraph]
            Graphs built from ground truth SWC files.
        fragment_graphs : Dict[str, FragmentsGraph]
            Graphs built from skeletons obtained from a segmentation.
        """
        # Initializations
        fragments_dir = os.path.join(self.output_dir, "fragments")
        util.mkdir(fragments_dir, delete=True)

        # Main
        for key, graph in gt_graphs.items():
            # Create zip writer
            zip_path = os.path.join(fragments_dir, f"{graph.name}.zip")
            zip_writer = ZipFile(zip_path, "a")

            # Save skeletons
            graph.to_zipped_swc(zip_writer)
            self.save_intersecting_fragments(
                graph, fragment_graphs, zip_writer
            )

    @staticmethod
    def save_intersecting_fragments(gt_graph, fragment_graphs, zip_writer):
        """
        Saves SWC files for all fragment graphs whose label intersects with
        the given ground-truth graph.

        Parameters
        ----------
        gt_graph : LabeledGraph
            Graphs built from ground truth SWC files.
        fragment_graphs : Dict[str, FragmentGraph]
            Graphs built from skeletons obtained from a segmentation.
        zip_writer : zipfile.ZipFile
            Open ZIP file handle used to write fragments.
        """
        intersecting_labels = gt_graph.get_node_labels()
        for key, graph in fragment_graphs.items():
            if graph.label in intersecting_labels:
                graph.to_zipped_swc(zip_writer)

    def save_merge_results(self, gt_graphs, fragment_graphs, output_dir):
        """
        Saves all detected merge results, including skeletons, merge sites,
        and metadata.

        Parameters
        ----------
        gt_graphs : Dict[str, LabeledGraph]
            Graphs built from ground truth SWC files.
        fragment_graphs : Dict[str, FragmentsGraph]
            Graphs built from skeletons obtained from a segmentation.
        output_dir : str
            Directory that results are written to.
        """
        # Initialize a writer
        filename = f"{self.results_filename}-merged_fragments.zip"
        zip_path = os.path.join(output_dir, filename)
        util.rm_file(zip_path)
        zip_writer = ZipFile(zip_path, "a")

        # Save SWC files
        self.save_merge_sites(zip_writer)
        self.save_skeletons_with_merge(gt_graphs, fragment_graphs, zip_writer)
        zip_writer.close()

        # Save CSV report
        filename = f"{self.results_filename}-merge_sites.csv"
        path = os.path.join(self.output_dir, filename)
        self.metrics["# Merges"].merge_sites.to_csv(path, index=True)

    def save_merge_sites(self, zip_writer):
        """
        Saves merge site coordinates into a ZIP archive.

        Parameters
        ----------
        zip_writer : zipfile.ZipFile
            Open ZIP file handle used to write merge site data.
        """
        merge_sites = self.metrics["# Merges"].merge_sites
        for i in range(len(merge_sites)):
            filename = merge_sites.index[i]
            xyz = merge_sites["World"].iloc[i]
            util.to_zipped_point(zip_writer, filename, xyz)

    def save_skeletons_with_merge(
        self, gt_graphs, fragment_graphs, zip_writer
    ):
        """
        Saves ground truth and fragment skeletons containing merge sites into
        a ZIP archive.

        Parameters
        ----------
        gt_graphs : Dict[str, LabeledGraph]
            Graphs built from ground truth SWC files.
        fragment_graphs : Dict[str, FragmentsGraph]
            Graphs built from skeletons obtained from a predicted segmentation.
        zip_writer : zipfile.ZipFile
            Open ZIP file handle used to write SWC data.
        """
        # Save ground truth skeletons
        keys = self.metrics["# Merges"].merge_sites["GroundTruth_ID"]
        for key in keys.unique():
            gt_graphs[key].to_zipped_swc(zip_writer)

        # Save fragments
        for key in self.metrics["# Merges"].fragments_with_merge:
            fragment_graphs[key].to_zipped_swc(zip_writer)



"""
Created on Thu Oct 16 12:00:00 2025

@author: Anna Grim
@email: anna.grim@alleninstitute.org

Module for loading graph structures, labeling nodes, and handling label
management.

"""

from collections import deque
from concurrent.futures import (
    as_completed,
    ProcessPoolExecutor,
    ThreadPoolExecutor,
)
from tqdm import tqdm

import networkx as nx
import numpy as np

from segmentation_skeleton_metrics.data_handling import swc_loading
from segmentation_skeleton_metrics.data_handling.graph_classes import (
    FragmentGraph,
    LabeledGraph,
)
from segmentation_skeleton_metrics.utils import util


class DataLoader:
    """
    A class that loads ground truth and fragment graphs and provides tools for
    labeling ground truth graphs.
    """

    def __init__(
        self,
        label_handler,
        anisotropy=(1.0, 1.0, 1.0),
        use_anisotropy=False,
        verbose=True,
    ):
        """
        Instantiates a DataLoader object.

        Parameters
        ----------
        label_handler : LabelHander
            Handles mapping between raw segmentation labels and consolidated
            class IDs.
        anisotropy : Tuple[int], optional
            Image to physical coordinates scaling factors to account for the
            anisotropy of the microscope. Default is (1.0, 1.0, 1.0).
        use_anisotropy : bool, optional
            Indication of whether to apply the anisotropy to the coordinates
            from the fragment SWC files. Default is False.
        verbose : bool, optional
            Indication of whether to display a progress bar. Default is True.
        """
        # Instance attributes
        self.anisotropy = anisotropy
        self.label_handler = label_handler
        self.use_anisotropy = use_anisotropy
        self.verbose = verbose

    # --- Core Routines ---
    def load_groundtruth(self, swc_pointer, label_mask):
        """
        Loads ground truth graphs.

        Parameters
        ----------
        swc_pointer : str
            Pointer to ground truth SWC files.
        label_mask : ImageReader
            Predicted segmentation.

        Returns
        -------
        Dict[str, SkeletonGraph]
            Ground truth graphs.
        """
        if self.verbose:
            print("\n(1) Load Ground Truth")

        graph_loader = GraphLoader(
            anisotropy=self.anisotropy,
            is_groundtruth=True,
            label_handler=self.label_handler,
            label_mask=label_mask,
            use_anisotropy=False,
            verbose=self.verbose,
        )
        return graph_loader(swc_pointer)

    def load_fragments(self, swc_pointer, gt_graphs):
        """
        Loads fragment graphs (predicted skeletons).

        Parameters
        ----------
        swc_pointer : str
            Path or pointer to predicted SWC files.
        gt_graphs : Dict[str, SkeletonGraph]
            Ground truth graphs to extract node labels from.

        Returns
        -------
        Dict[str, SkeletonGraph] or None
            Fragment graphs or None.
        """
        if self.verbose:
            print("\n(2) Load Fragments")

        # Check if SWC pointer is provided
        if not swc_pointer:
            return None

        # Load fragments
        selected_ids = self.get_all_node_labels(gt_graphs)
        graph_loader = GraphLoader(
            anisotropy=self.anisotropy,
            is_groundtruth=False,
            label_handler=self.label_handler,
            selected_ids=selected_ids,
            use_anisotropy=self.use_anisotropy,
            verbose=self.verbose,
        )
        return graph_loader(swc_pointer)

    # --- Helpers ---
    def get_all_node_labels(self, graphs):
        """
        Gets the set of unique node labels across all given graphs.

        Parameters
        ----------
        graphs : Dict[str, SkeletonGraph]
            Graph to be searched.

        Returns
        -------
        node_labels : Set[int]
            Unique node labels across all graphs.
        """
        node_labels = set()
        for graph in graphs.values():
            node_labels |= self.label_handler.get_node_labels(graph)
        return node_labels


class GraphLoader:
    """
    A class that builds graphs from SWC files.
    """

    def __init__(
        self,
        anisotropy=(1.0, 1.0, 1.0),
        fix_label_misalignments=True,
        is_groundtruth=False,
        label_handler=None,
        label_mask=None,
        selected_ids=None,
        use_anisotropy=True,
        verbose=True,
    ):
        """
        Instantiates a GraphLoader object.

        Parameters
        ----------
        anisotropy : Tuple[int], optional
            Image to physical coordinates scaling factors to account for the
            anisotropy of the microscope. Default is (1.0, 1.0, 1.0).
        fix_label_misalignments : bool, optional
            Indication of whether to fix misalignments between skeletons and
            segmentation mask. Default is True.
        is_groundtruth : bool, optional
            Indication of whether this graph corresponds to a ground truth
            tracing. Default is False.
        label_mask : ImageReader, optional
            Predicted segmentation mask.
        selected_ids : Set[int], optional
            Only SWC files with a name contained in this set are read. Default
            is None.
        use_anisotropy : bool, optional
            Indication of whether coordinates in SWC files should be converted
            from physical to image coordinates using the given anisotropy.
            Default is True.
        verbose : bool, optional
            Indication of whether to display a progress bar. Default is True.
        """
        # Instance attributes
        self.anisotropy = np.array(anisotropy)
        self.fix_label_misalignments = fix_label_misalignments
        self.is_groundtruth = is_groundtruth
        self.label_handler = label_handler
        self.label_mask = label_mask
        self.use_anisotropy = use_anisotropy
        self.verbose = verbose

        # Reader
        self.swc_reader = swc_loading.Reader(
            selected_ids=selected_ids, verbose=verbose
        )

    def __call__(self, swc_pointer):
        """
        Builds a graphs by reading SWC files to extract content to load into a
        SkeletonGraph object. Nodes are labeled if a label_mask is provided.

        Parameters
        ----------
        swc_pointer : Any
            Object that points to SWC files to be read.

        Returns
        -------
        graphs : Dict[str, SkeletonGraph]
            Dictionary where the keys are unique identifiers (i.e. filenames
            of SWC files) and values are the corresponding SkeletonGraph.
        """
        graphs = self._build_graphs_from_swcs(swc_pointer)
        if self.label_mask:
            for name in self.iterator(graphs, desc="Label Graphs"):
                self._label_graph(graphs[name])
                self._fix_label_misalignments(graphs[name])
        return graphs

    # --- Build Graphs ---
    def _build_graphs_from_swcs(self, swc_pointer):
        """
        Builds graphs by reading SWC files to extract content which is then
        loaded into a custom SkeletonGraph object.

        Parameters
        ----------
        swc_pointer : Any
            Object that points to SWC files to be read.

        Returns
        -------
        graphs : Dict[str, SkeletonGraph]
            Dictionary where the keys are unique identifiers (i.e. SWC
            filenames) and values are the corresponding SkeletonGraphs.
        """
        # Initializations
        swc_dicts = self.swc_reader(swc_pointer)
        pbar = self.manual_progress_bar(len(swc_dicts), desc="Build Graphs")

        # Main
        graphs = dict()
        if len(swc_dicts) > 10**4:
            while len(swc_dicts) > 0:
                swc_dict = swc_dicts.pop()
                graphs.update(self.to_graph(swc_dict))
                if self.verbose:
                    pbar.update(1)
        else:
            with ProcessPoolExecutor(max_workers=1) as executor:
                # Assign processes
                processes = list()
                while len(swc_dicts) > 0:
                    swc_dict = swc_dicts.pop()
                    processes.append(executor.submit(self.to_graph, swc_dict))

                # Store results
                for process in as_completed(processes):
                    graphs.update(process.result())
                    if self.verbose:
                        pbar.update(1)
        return graphs

    def to_graph(self, swc_dict):
        """
        Builds a graph from a dictionary that contains the contents of an SWC
        file.

        Parameters
        ----------
        swc_dict : dict
            Dictionary whose keys and values are the attribute names and
            values from an SWC file.

        Returns
        -------
        Dict[str, SkeletonGraph]
            Graph built from an SWC file.
        """
        # Initialize graph
        graph = self._init_graph(swc_dict)

        # Build graph structure
        id_lookup = dict()
        for i, id_i in enumerate(swc_dict["id"]):
            id_lookup[id_i] = i
            if swc_dict["pid"][i] != -1:
                parent = id_lookup[swc_dict["pid"][i]]
                graph.add_edge(i, parent)
                graph.run_length += graph.dist(i, parent)
        graph.prune_branches()

        # Apply voxel coordinate conversion (if applicable)
        if self.use_anisotropy:
            graph.voxels = (graph.voxels / self.anisotropy).astype(int)
            graph.voxels[:, [0, 2]] = graph.voxels[:, [2, 0]]
        return {graph.name: graph}

    def _init_graph(self, swc_dict):
        """
        Initializes and returns a graph object from a parsed SWC dictionary.

        Parameters
        ----------
        swc_dict : dict
            Dictionary whose keys and values are the attribute names and
            values from an SWC file.

        Returns
        -------
        graph : SkeletonGraph
            An initialized LabeledGraph or FragmentGraph instance with voxel
            data, filename, and node count set.
        """
        # Instantiate graph
        if self.is_groundtruth:
            graph = LabeledGraph(
                anisotropy=self.anisotropy, name=swc_dict["swc_name"]
            )
        else:
            segment_id = util.get_segment_id(swc_dict["swc_name"])
            label = self.label_handler.get(segment_id)
            graph = FragmentGraph(
                anisotropy=self.anisotropy,
                name=swc_dict["swc_name"],
                label=label,
                segment_id=segment_id,
            )

        # Set class attributes
        graph.init_voxels(swc_dict["voxel"])
        graph.set_filename(swc_dict["swc_name"] + ".swc")
        graph.set_nodes(len(swc_dict["id"]))
        return graph

    # --- Label Graphs ---
    def _label_graph(self, graph):
        """
        Assigns labels to graph nodes by indexing a segmentation mask using
        each node’s voxel coordinates.

        Parameters
        ----------
        graph : LabeledGraph
            Graph to be labeled.
        """
        with ThreadPoolExecutor() as executor:
            # Assign threads
            batch = set()
            threads = list()
            visited = set()
            for i, j in nx.dfs_edges(graph):
                # Check whether to start new batch
                if len(batch) == 0:
                    root = i
                    batch.add(i)
                    visited.add(i)

                # Check whether to submit batch
                is_node_far = graph.dist(root, j) > 128
                is_batch_full = len(batch) >= 128
                if is_node_far or is_batch_full:
                    threads.append(
                        executor.submit(self.get_patch_labels, graph, batch)
                    )
                    batch = set()

                # Visit j
                if j not in visited:
                    batch.add(j)
                    visited.add(j)
                    if len(batch) == 1:
                        root = j

            # Submit last batch
            threads.append(
                executor.submit(self.get_patch_labels, graph, batch)
            )

            # Store results
            graph.init_node_labels()
            for thread in as_completed(threads):
                node_to_label = thread.result()
                for i, label in node_to_label.items():
                    graph.node_labels[i] = label

    def get_patch_labels(self, graph, nodes):
        """
        Gets the segment labels for a given set of nodes within a specified
        patch of the label mask.

        Parameters
        ----------
        graph : str
            Unique identifier of graph to be labeled.
        nodes : List[int]
            Node IDs for which the labels are to be retrieved.

        Returns
        -------
        node_to_label : Dict[int, int]
            Dictionary that maps node IDs to their respective labels.
        """
        bbox = graph.get_bbox(nodes)
        label_patch = self.label_mask.read_with_bbox(bbox)
        node_to_label = dict()
        for i in nodes:
            voxel = self.to_local_voxels(graph, i, bbox["min"])
            label = self.label_handler.get(label_patch[voxel])
            node_to_label[i] = label
        return node_to_label

    def _fix_label_misalignments(self, graph):
        """
        Adjusts misalignments between the labeled graph and segmentation.

        Parameters
        ----------
        graph : LabeledGraph
            Graph to be searched.
        """
        if self.fix_label_misalignments:
            visited = set()
            for i, j in deque(nx.dfs_edges(graph)):
                # Check whether to visit edge
                if frozenset({i, j}) in visited:
                    continue

                # Visit edge
                if int(graph.node_labels[j]) == 0:
                    GraphLoader.check_misalignment(graph, visited, i, j)
                visited.add(frozenset({i, j}))

    # --- Helpers ---
    def iterator(self, iterator, desc=""):
        """
        Gets an iterator that optionally displays a progress bar.

        Parameters
        ----------
        iterator : iterable
            Object to be iterated over.
        desc : str, optional
            Text to display on progress bar. Default is an empty string.

        Returns
        -------
        tqdm.tqdm
            Iterator that is optionally wrapped in a progress bar.
        """
        return tqdm(iterator, desc=desc) if self.verbose else iterator

    def manual_progress_bar(self, total, desc=""):
        """
        Gets progress bar that needs to be updated manually.

        Parameters
        ----------
        total : int
            Size of progress bar.
        desc : str, optional
            Text to be displayed on progress bar. Default is an empty string.

        Returns
        -------
        tqdm.tqdm
            Iterator that is optionally wrapped in a progress bar.
        """
        return tqdm(total=total, desc=desc) if self.verbose else None

    def to_local_voxels(self, graph, i, offset):
        """
        Converts from global to local voxel coordinates.

        Parameters
        ----------
        graph : SkeletonGraph
            Graph object containing node voxel coordinates.
        i : int
            Node ID of voxel coordinate to be converted.
        offset : ArrayLike
            Offset to subtract from the global voxel coordinate to get the
            local coordinate.

        Returns
        -------
        Tuple[int]
            Local voxel coordinate after subtracting the offset.
        """
        voxel = np.array(graph.voxels[i])
        offset = np.array(offset)
        return tuple(voxel - offset)

    @staticmethod
    def check_misalignment(graph, visited_edges, nb, root):
        """
        Determines whether zero-valued label corresponds to a misalignment
        between the graph and segmentation mask.

        Parameters
        ----------
        graph : LabeledGraph
            Graph that represents a ground truth neuron.
        visited_edges : List[Frozenset[int]]
            List of edges in "graph" that have been visited.
        nb : int
            Neighbor of "root".
        root : int
            Node where possible misalignment starts (i.e. zero-valued label).
        """
        # Search graph
        label_collisions = set()
        queue = deque([root])
        visited = set()
        while len(queue) > 0:
            # Visit node
            j = queue.popleft()
            label_j = int(graph.node_labels[j])
            if label_j != 0:
                label_collisions.add(label_j)
            visited.add(j)

            # Update queue
            if label_j == 0:
                for k in graph.neighbors(j):
                    if k not in visited:
                        if frozenset({j, k}) not in visited_edges or k == nb:
                            queue.append(k)
                            visited_edges.add(frozenset({j, k}))

        # Upd zero nodes
        if len(label_collisions) == 1:
            label = label_collisions.pop()
            graph.update_node_labels(visited, label)


class LabelHandler:
    """
    Handles mapping between raw segmentation labels and consolidated class IDs.

    The class is designed to manage cases where multiple segment IDs are merged
    into a single equivalence class. It supports:
      - Building mappings from a file of pairwise segment connections.
      - Mapping individual labels to class IDs.
      - Retrieving all labels belonging to a given class.
      - Enforcing constraints on which labels are considered valid.

    Attributes
    ----------
    mapping : Dict[int, int]
        Maps a raw label (segment ID) to its class ID.
    inverse_mapping : Dict[int, Set[int]]
        Maps a class ID back to the set of raw labels it contains.
    processed_labels : Set[int]
        Labels that have been processed during initialization.
    valid_labels : Set[int]
        Labels that are allowed to be assigned (after filtering).
    """

    def __init__(self, connections_path=None, valid_labels=set()):
        """
        Instantiates a LabelHandler object and optionally builds label
        mappings.

        Parameters
        ----------
        connections_path : str, optional
            Path to file containing pairs of segment IDs that were merged.
            Default is None.
        valid_labels : Set[int], optional
            Subset of labels that are considered to be valid. This argument
            accounts for segments removed due to filtering. Default is an
            empty set.
        """
        self.mapping = dict()  # Maps label to equivalent class id
        self.inverse_mapping = dict()  # Maps class id to list of labels
        self.processed_labels = set()
        self.valid_labels = valid_labels
        if connections_path:
            self.init_mappings(connections_path)

    # --- Constructor Helpers ---
    def init_mappings(self, connections_path):
        """
        Initializes dictionaries that map between segment IDs and equivalent
        class IDS.

        Parameters
        ----------
        connections_path : str
            Path to file containing pairs of segment IDs that were merged.
        """
        self.mapping = {0: 0}
        self.inverse_mapping = {0: [0]}
        labels_graph = self.build_labels_graph(connections_path)
        for i, labels in enumerate(nx.connected_components(labels_graph)):
            class_id = i + 1
            self.inverse_mapping[class_id] = set()
            for label in labels:
                self.mapping[label] = class_id
                self.inverse_mapping[class_id].add(label)

    def build_labels_graph(self, connections_path):
        """
        Builds a graph of labels from valid labels and merge connections.
        Nodes correspond to "self.valid_labels", and edges are added between
        labels that were merged according to the file.

        Parameters
        ----------
        connections_path : str
            Path to a text file containing merge connections. Each line should
            specify a pair of segment IDs separated by a comma.

        Returns
        -------
        labels_graph : networkx.Graph
            Graph with nodes that represent labels and edges are based on the
            connections read from the "connections_path".
        """
        # Initializations
        assert self.valid_labels is not None, "Must provide valid labels!"
        labels_graph = nx.Graph()
        labels_graph.add_nodes_from(self.valid_labels)

        # Main
        for line in util.read_txt(connections_path).splitlines():
            ids = line.split(",")
            id_1 = util.get_segment_id(ids[0])
            id_2 = util.get_segment_id(ids[1])
            labels_graph.add_edge(id_1, id_2)
        return labels_graph

    # --- Core Routines ---
    def get(self, label):
        """
        Maps a raw label to its class ID.

        Parameters
        ----------
        label : int
            Raw label (segment ID) to be mapped.

        Returns
        -------
        int
            Class ID corresponding to the label.
        """
        if self.use_mapping():
            return self.mapping.get(label, 0)
        elif self.valid_labels:
            return 0 if label not in self.valid_labels else label
        return label

    def get_class(self, label):
        """
        Gets all raw labels associated with a class ID.

        Parameters
        ----------
        label : int
            Class ID or raw label.

        Returns
        -------
        List[int]
            Labels corresponding to the class.
        """
        return self.inverse_mapping[label] if self.use_mapping() else [label]

    def use_mapping(self):
        """
        Checks whether mappings have been initialized.

        Returns
        -------
        bool
            True if mappings are active, False otherwise.
        """
        return True if len(self.mapping) > 0 else False

    # --- Helpers ---
    def get_node_labels(self, graph):
        """
        Gets the set of unique node labels from the given graph.

        Parameters
        ----------
        graph : LabeledGraph
            Graph from which to retrieve the node labels.

        Returns
        -------
        labels : Set[int]
            Labels corresponding to nodes in the graph identified by "key".
        """
        labels = graph.get_node_labels()
        if self.use_mapping():
            labels = set().union(*(self.inverse_mapping[u] for u in labels))
        return labels



"""
Created on Wed June 5 16:00:00 2023

@author: Anna Grim
@email: anna.grim@alleninstitute.org

Routines for working with SWC files. An SWC file is a text-based file format
used to represent the directed graphical structure of a neuron. It contains a
series of nodes such that each has the following attributes:
    "id" (int): node ID
    "type" (int): node type (e.g. soma, axon, dendrite)
    "x" (float): x coordinate
    "y" (float): y coordinate
    "z" (float): z coordinate
    "pid" (int): node ID of parent

Note: Each uncommented line in an SWC file corresponds to a node and contains
      these attributes in the same order.
"""

from botocore import UNSIGNED
from botocore.client import Config
from collections import deque
from concurrent.futures import (
    as_completed,
    ProcessPoolExecutor,
    ThreadPoolExecutor,
)
from google.cloud import storage
from io import BytesIO
from tqdm import tqdm
from zipfile import ZipFile

import boto3
import numpy as np
import os

from segmentation_skeleton_metrics.utils import util


class Reader:
    """
    Class that reads SWC files stored in a (1) local directory, (2) local ZIP
    archive, and (3) local directory of ZIP archives.
    """

    def __init__(self, selected_ids=None, verbose=True):
        """
        Initializes a Reader object that reads SWC files.

        Parameters
        ----------
        selected_ids : Set[int], optional
            Only SWC files with an swc_id contained in this set are read.
            Default is None.
        verbose : bool, optional
            Indication of whether to display a progress bar. Default is True.
        """
        self.selected_ids = selected_ids or set()
        self.verbose = verbose

    # --- Read Data ---
    def __call__(self, swc_pointer):
        """
        Loads SWC files based on the type pointer provided.

        Parameters
        ----------
        swc_pointer : str
            Object that points to SWC files to be read, must be one of:
                - file_path: Path to single SWC file
                - dir_path: Path to local directory with SWC files
                - zip_path: Path to local ZIP with SWC files
                - zip_dir_path: Path to local directory of ZIPs with SWC files
                - s3_dir_path: Path to S3 prefix with SWC files
                - gcs_dir_path: Path to GCS prefix with SWC files
                - gcs_zip_dir_path: Path to GCS prefix with ZIPs of SWC files

        Returns
        -------
        Deque[dict]
            Dictionaries whose keys and values are the attribute names and
            values from the SWC files. Each dictionary contains the following:
            items:
                - "id": unique identifier of each node in an SWC file.
                - "pid": parent ID of each node.
                - "radius": radius value corresponding to each node.
                - "xyz": coordinate corresponding to each node.
                - "filename": filename of SWC file
                - "swc_id": name of SWC file, minus the ".swc".
        """
        # Directory containing...
        if os.path.isdir(swc_pointer):
            # Local ZIP archives with SWC files
            paths = util.list_paths(swc_pointer, extension=".zip")
            if len(paths) > 0:
                return self.read_from_zips(swc_pointer)

            # Local SWC files
            paths = util.read_paths(swc_pointer, extension=".swc")
            if len(paths) > 0:
                return self.read_from_paths(paths)

            raise Exception("Directory is Invalid!")

        # Path to...
        if isinstance(swc_pointer, str):
            # Cloud GCS storage
            if util.is_gcs_path(swc_pointer):
                return self.read_from_gcs(swc_pointer)

            # Cloud S3 storage
            if util.is_s3_path(swc_pointer):
                return self.read_from_s3(swc_pointer)

            # Local ZIP archive with SWC files
            if swc_pointer.endswith(".zip"):
                return self.read_from_zip(swc_pointer)

            # Local path to single SWC file
            if swc_pointer.endswith(".swc"):
                return self.read_from_path(swc_pointer)

            raise Exception("Path is Invalid!")

        raise Exception("SWC Pointer is Invalid!")

    def read_from_path(self, path):
        """
        Reads a single SWC file.

        Paramters
        ---------
        path : str
            Path to SWC file.

        Returns
        -------
        dict
            Dictionary whose keys and values are the attribute names and
            values from an SWC file.
        """
        content = util.read_txt(path)
        filename = os.path.basename(path)
        if self.confirm_read(filename):
            return self.parse(content, filename)
        else:
            return None

    def read_from_paths(self, paths):
        """
        Reads SWC files given a list of paths.

        Paramters
        ---------
        swc_paths : List[str]
            Paths to SWC files.

        Returns
        -------
        swc_dicts : Deque[dict]
            Dictionaries whose keys and values are the attribute names and
            values from an SWC file.
        """
        with ThreadPoolExecutor() as executor:
            # Assign threads
            threads = list()
            for path in paths:
                filename = os.path.basename(path)
                if self.confirm_read(filename):
                    threads.append(executor.submit(self.read_from_path, path))

            # Store results
            swc_dicts = deque()
            pbar = self.manual_progress_bar(len(paths))
            for thread in as_completed(threads):
                result = thread.result()
                if result:
                    swc_dicts.append(result)
                if self.verbose:
                    pbar.update(1)
        return swc_dicts

    def read_from_zips(self, zip_dir):
        """
        Processes a directory containing ZIP archives with SWC files.

        Parameters
        ----------
        zip_dir : str
            Path to directory containing ZIP archives with SWC files.

        Returns
        -------
        swc_dicts : Deque[dict]
            Dictionaries whose keys and values are the attribute names and
            values from an SWC file.
        """
        with ProcessPoolExecutor() as executor:
            # Assign threads
            processes = list()
            zip_names = [f for f in os.listdir(zip_dir) if f.endswith(".zip")]
            for name in zip_names:
                zip_path = os.path.join(zip_dir, name)
                processes.append(executor.submit(self.read_from_zip, zip_path))

            # Store results
            swc_dicts = deque()
            pbar = self.manual_progress_bar(len(zip_names))
            for process in as_completed(processes):
                swc_dicts.extend(process.result())
                if self.verbose:
                    pbar.update(1)
        return swc_dicts

    def read_from_zip(self, zip_path):
        """
        Reads SWC files from a ZIP archive.

        Paramters
        ---------
        zip_path : str
            Path to ZIP archive.

        Returns
        -------
        swc_dicts : Deque[dict]
            Dictionaries whose keys and values are the attribute names and
            values from an SWC file.
        """
        with ThreadPoolExecutor() as executor:
            # Assign threads
            threads = list()
            zipfile = ZipFile(zip_path, "r")
            filenames = [f for f in zipfile.namelist() if f.endswith(".swc")]
            for filename in filenames:
                if self.confirm_read(filename):
                    threads.append(
                        executor.submit(
                            self.read_from_zipped_file, zipfile, filename
                        )
                    )

            # Store results
            swc_dicts = deque()
            for thread in as_completed(threads):
                result = thread.result()
                if result:
                    swc_dicts.append(result)
        return swc_dicts

    def read_from_zipped_file(self, zipfile, path):
        """
        Reads an SWC file stored in a ZIP archive.

        Parameters
        ----------
        zip_file : ZipFile
            ZIP archive containing SWC files.
        path : str
            Path to SWC file.

        Returns
        -------
        dict
            Dictionary whose keys and values are the attribute names and
            values from an SWC file.
        """
        content = util.read_zip(zipfile, path).splitlines()
        filename = os.path.basename(path)
        return self.parse(content, filename)

    def read_from_gcs(self, gcs_path):
        """
        Reads SWC files stored in a GCS bucket.

        Parameters
        ----------
        gcs_path : str
            Path to location in a GCS bucket that the SWC files are stored.
            The path must be in the format "gs://{bucket_name}/{prefix}",
            where "prefix" is a path to a directory containing SWC files or
            ZIP archives containing SWC files

        Returns
        -------
        Dequeue[dict]
            Dictionaries whose keys and values are the attribute
            names and values from an SWC file.
        """
        # List filenames
        bucket_name, prefix = util.parse_cloud_path(gcs_path)
        swc_paths = util.list_gcs_filenames(bucket_name, prefix, ".swc")
        zip_paths = util.list_gcs_filenames(bucket_name, prefix, ".zip")

        # Call reader
        if len(swc_paths) > 0:
            return self.read_from_gcs_swcs(bucket_name, swc_paths)
        if len(zip_paths) > 0:
            return self.read_from_gcs_zips(bucket_name, zip_paths)

        # Error
        raise Exception(f"GCS Pointer is invalid {gcs_path}")

    def read_from_gcs_swcs(self, bucket_name, swc_paths):
        """
        Reads SWC files stored in a GCS bucket.

        Parameters
        ----------
        bucket_name : str
            Name of bucket containing SWC files.
        swc_paths : List[str]
            List of paths to SWC files to be read.

        Returns
        -------
        swc_dicts : Dequeue[dict]
            Dictionaries whose keys and values are the attribute
            names and values from an SWC file.
        """
        with ThreadPoolExecutor() as executor:
            # Assign threads
            threads = list()
            for path in swc_paths:
                threads.append(
                    executor.submit(self.read_from_gcs_swc, bucket_name, path)
                )

            # Store results
            swc_dicts = deque()
            pbar = self.manual_progress_bar(len(threads))
            for thread in as_completed(threads):
                result = thread.result()
                if result:
                    swc_dicts.append(result)
                if self.verbose:
                    pbar.update(1)
        return swc_dicts

    def read_from_gcs_swc(self, bucket_name, path):
        """
        Reads a single SWC file stored in a GCS bucket.

        Parameters
        ----------
        bucket_name : str
            Name of bucket containing SWC file.
        path : List[str]
            Path to SWC file to be read.

        Returns
        -------
        dict
            Dictionary whose keys and values are the attribute names and
            values from an SWC file.
        """
        # Initialize cloud reader
        client = storage.Client()
        bucket = client.bucket(bucket_name)
        blob = bucket.blob(path)

        # Parse swc contents
        content = blob.download_as_text().splitlines()
        filename = os.path.basename(path)
        return self.parse(content, filename)

    def read_from_gcs_zips(self, bucket_name, zip_paths):
        """
        Reads SWC files stored in a ZIP archives stored in a GCS bucket.

        Parameters
        ----------
        bucket_name : str
            Name of bucket containing SWC files.
        zip_paths : List[str]
            List of paths to ZIP archives containing SWC files to be read.

        Returns
        -------
        swc_dicts : Dequeue[dict]
            Dictionaries whose keys and values are the attribute names and
            values from an SWC file.
        """
        swc_dicts = deque()
        for zip_path in self.iterator(zip_paths):
            swc_dicts.extend(self.read_from_gcs_zip(bucket_name, zip_path))
        return swc_dicts

    def read_from_gcs_zip(self, bucket_name, path):
        """
        Reads SWC files stored in a ZIP archive downloaded from a GCS
        bucket.

        Parameters
        ----------
        bucket_name : str
            Name of bucket containing SWC files.
        path : str
            Path to ZIP archive containing SWC files to be read.

        Returns
        -------
        swc_dicts : Dequeue[dict]
            Dictionaries whose keys and values are the attribute names and
            values from an SWC file.
        """
        # Initialize cloud reader
        client = storage.Client()
        bucket = client.bucket(bucket_name)

        # Parse Zip
        swc_dicts = deque()
        zip_content = bucket.blob(path).download_as_bytes()
        with ZipFile(BytesIO(zip_content), "r") as zip_file:
            with ThreadPoolExecutor() as executor:
                # Assign threads
                threads = list()
                for filename in zip_file.namelist():
                    if self.confirm_read(filename):
                        threads.append(
                            executor.submit(
                                self.read_from_zipped_file, zip_file, filename
                            )
                        )

                # Process results
                for thread in as_completed(threads):
                    result = thread.result()
                    if result:
                        swc_dicts.append(result)
        return swc_dicts

    def read_from_s3(self, s3_path):
        """
        Reads and parses SWC files from an S3 directory.

        Parameters
        ----------
        s3_path : str
            Path to a directory in an S3 bucket containing SWC files or ZIPs
            of SWC files to be read.

        Returns
        -------
        swc_dicts : Dequeue[dict]
            Dictionaries whose keys and values are the attribute names and
            values from an SWC file.
        """
        # List filenames
        bucket_name, prefix = util.parse_cloud_path(s3_path)
        swc_paths = util.list_s3_filenames(bucket_name, prefix, ".swc")
        zip_paths = util.list_s3_filenames(bucket_name, prefix, ".zip")

        # Call reader
        if len(swc_paths) > 0:
            return self.read_from_s3_swcs(bucket_name, swc_paths)
        if len(zip_paths) > 0:
            return self.read_from_s3_zips(bucket_name, zip_paths)

    def read_from_s3_swcs(self, bucket_name, swc_paths):
        """
        Reads and parses SWC files from an S3 directory.

        Parameters
        ----------
        s3_path : str
            Path to a directory in an S3 bucket containing SWC files to be
            read.

        Returns
        -------
        swc_dicts : Dequeue[dict]
            Dictionaries whose keys and values are the attribute names and
            values from an SWC file.
        """
        swc_dicts = deque()
        for path in self.iterator(swc_paths):
            content = util.read_txt(bucket_name, path).splitlines()
            filename = os.path.basename(path)
            result = self.parse(content, filename)
            if result:
                swc_dicts.append(result)
        return swc_dicts

    def read_from_s3_zips(self, bucket_name, zip_paths):
        """
        Reads SWC files stored in a list of ZIP archives stored in an S3
        bucket.

        Parameters
        ----------
        bucket_name : str
            Name of bucket containing SWC files.
        zip_paths : str
            Path to ZIP archive containing SWC files to be read.

        Returns
        -------
        swc_dicts : Dequeue[dict]
            Dictionaries whose keys and values are the attribute names and
            values from an SWC file.
        """
        with ProcessPoolExecutor() as executor:
            # Submit processes
            processes = list()
            for zip_path in zip_paths:
                processes.append(
                    executor.submit(
                        self.read_from_s3_zip, bucket_name, zip_path
                    )
                )

            # Store results
            swc_dicts = deque()
            pbar = self.manual_progress_bar(len(processes))
            for process in as_completed(processes):
                result = process.result()
                if result:
                    swc_dicts.extend(result)
                if self.verbose:
                    pbar.update(1)
        return swc_dicts

    def read_from_s3_zip(self, bucket_name, path):
        """
        Reads SWC files stored in a ZIP archive downloaded from an S3
        bucket.

        Parameters
        ----------
        bucket_name : str
            Name of bucket containing SWC files.
        path : str
            Path to ZIP archive containing SWC files to be read.

        Returns
        -------
        swc_dicts : Dequeue[dict]
            Dictionaries whose keys and values are the attribute names and
            values from an SWC file.
        """
        # Initialize cloud reader
        s3 = boto3.client("s3", config=Config(signature_version=UNSIGNED))
        zip_obj = s3.get_object(Bucket=bucket_name, Key=path)
        zip_content = zip_obj["Body"].read()

        # Parse ZIP
        with ZipFile(BytesIO(zip_content), "r") as zip_file:
            with ThreadPoolExecutor() as executor:
                # Assign threads for reading files
                threads = [
                    executor.submit(
                        self.read_from_zipped_file, zip_file, filename
                    )
                    for filename in zip_file.namelist()
                    if self.confirm_read(filename)
                ]

                # Collect results
                swc_dicts = deque()
                pbar = self.manual_progress_bar(len(threads))
                for thread in as_completed(threads):
                    result = thread.result()
                    if result:
                        swc_dicts.append(result)
                    if self.verbose:
                        pbar.update(1)
        return swc_dicts

    def confirm_read(self, filename):
        """
        Checks whether the swc_id corresponding to the given filename is
        contained in the attribute "selected_ids".

        Parameters
        ----------
        filename : str
            Name of SWC file to be checked.

        Returns
        -------
        bool
            Indication of whether to read SWC file.
        """
        if self.selected_ids:
            segment_id = util.get_segment_id(filename)
            return True if segment_id in self.selected_ids else False
        else:
            return True

    # -- Process Text ---
    def iterator(self, iterator):
        """
        Gets an iterator that optionally displays a progress bar.

        Parameters
        ----------
        iterator : iterable
            Object to be iterated over.

        Returns
        -------
        tqdm.tqdm
            Iterator that is optionally wrapped in a progress bar.
        """
        return tqdm(iterator, desc="Read SWCs") if self.verbose else iterator

    def manual_progress_bar(self, total):
        """
        Gets progress bar that needs to be updated manually.

        Parameters
        ----------
        total : int
            Size of progress bar.

        Returns
        -------
        tqdm.tqdm
            Iterator that is optionally wrapped in a progress bar.
        """
        return tqdm(total=total, desc="Read SWCs") if self.verbose else None

    def parse(self, content, filename):
        """
        Parses an SWC file to extract the content which is stored in a dict.

        Parameters
        ----------
        content : List[str]
            List of strings such that each is a line from an SWC file.

        Returns
        -------
        dict
            Dictionary whose keys and values are the attribute names and
            values from an SWC file.
        """
        # Initializations
        swc_name, _ = os.path.splitext(filename)
        content, offset = self.process_content(content)
        if len(content) > 30:
            swc_dict = {
                "id": np.zeros((len(content)), dtype=int),
                "pid": np.zeros((len(content)), dtype=int),
                "voxel": np.zeros((len(content), 3), dtype=np.int32),
                "swc_name": swc_name,
            }

            # Parse content
            for i, line in enumerate(content):
                parts = line.split()
                swc_dict["id"][i] = parts[0]
                swc_dict["pid"][i] = parts[-1]
                swc_dict["voxel"][i] = self.read_coordinate(parts[2:5], offset)
            return swc_dict
        else:
            return None

    def process_content(self, content):
        """
        Processes lines of text from an SWC file, extracting an offset
        value and returning the remaining content starting from the line
        immediately after the last commented line.

        Parameters
        ----------
        content : List[str]
            List of strings such that each is a line from an SWC file.

        Returns
        -------
        content : List[str]
            Lines from an SWC file after comments.
        offset : Tuple[int]
            Offset used to shift coordinate.
        """
        offset = (0, 0, 0)
        for i, line in enumerate(content):
            if line.startswith("# OFFSET"):
                parts = line.split()
                offset = self.read_coordinate(parts[2:5])
            if not line.startswith("#") and len(line.strip()) > 0:
                return content[i:], offset

    def read_coordinate(self, xyz_str, offset=(0, 0, 0)):
        """
        Reads a coordinate from a string and converts it to voxel coordinates.

        Parameters
        ----------
        xyz_str : str
            Coordinate stored as a string.
        offset : Tuple[int]
            Offset of coordinates in SWC file. Default is (0, 0, 0).

        Returns
        -------
        Tuple[int]
            xyz coordinates of an entry from an SWC file.
        """
        return [float(xyz_str[i]) + offset[i] for i in range(3)]



# -*- coding: utf-8 -*-
"""
Created on Wed Dec 21 19:00:00 2022

@author: Anna Grim
@email: anna.grim@alleninstitute.org

Code for helper routines.

"""

from botocore import UNSIGNED
from botocore.client import Config
from collections import deque
from random import sample
from google.cloud import storage
from io import BytesIO, StringIO
from zipfile import ZipFile

import boto3
import os
import pandas as pd
import shutil


# -- OS Utils ---
def mkdir(path, delete=False):
    """
    Creates a directory at the given path.

    Parameters
    ----------
    path : str
        Path of directory to be created.
    delete : bool, optional
        Indication of whether to delete directory at path if it already
        exists. Default is False.
    """
    if delete:
        rmdir(path)

    os.makedirs(path, exist_ok=True)


def rmdir(path):
    """
    Removes the given directory and all of its subdirectories.

    Parameters
    ----------
    path : str
        Path to directory to be removed if it exists.
    """
    if os.path.exists(path):
        shutil.rmtree(path)


def rm_file(path):
    """
    Removes the file at the given path.

    Parameters
    ----------
    path : str
        Path to file to be removed.
    """
    if os.path.exists(path):
        os.remove(path)


def list_dir(directory, extension=None):
    """
    Lists filenames in the given directory. If "extension" is provided,
    filenames ending with the given extension are returned.

    Parameters
    ----------
    directory : str
        Path to directory to be searched.
    extension : str, optional
       Extension of filenames to be returned. Default is None.

    Returns
    -------
    List[str]
        Filenames in the given directory.
    """
    if extension is None:
        return [f for f in os.listdir(directory)]
    else:
        return [f for f in os.listdir(directory) if f.endswith(extension)]


def list_files_in_zip(zip_content):
    """
    Lists all files in a zip file stored in a GCS bucket.

    Parameters
    ----------
    zip_content : str
        Content stored in a ZIP archive in the form of a string of bytes.

    Returns
    -------
    List[str]
        Filenames in a ZIP archive file.
    """
    with ZipFile(BytesIO(zip_content), "r") as zip_file:
        return zip_file.namelist()


def list_paths(directory, extension=None):
    """
    Lists paths of files in the given directory. If "extension" is provided,
    filenames ending with the given extension are returned.

    Parameters
    ----------
    directory : str
        Path to directory to be searched.
    extension : str, optional
        Extension of filenames to be returned. Default is None.

    Returns
    -------
    paths : List[str]
        Paths of files in the given directory.
    """
    paths = list()
    for f in list_dir(directory, extension=extension):
        paths.append(os.path.join(directory, f))
    return paths


# --- IO Utils ---
def read_json(path):
    """
    Reads JSON file located at the given path.

    Parameters
    ----------
    path : str
        Path to JSON file to be read.

    Returns
    -------
    dict
        Contents of JSON file.
    """
    return pd.read_json(path, storage_options={"anon": True}, typ="series")


def read_txt(path):
    """
    Reads txt file at the given path.

    Parameters
    ----------
    path : str
        Path to txt file.

    Returns
    -------
    List[str]
        Lines from the txt file.
    """
    if is_s3_path(path):
        return read_txt_from_s3(path)
    elif is_gcs_path(path):
        return read_txt_from_gcs(path)
    else:
        with open(path, "r") as f:
            return f.read()


def read_zip(zip_file, path):
    """
    Reads txt file contained in the given ZIP archive.

    Parameters
    ----------
    zip_file : ZipFile
        ZIP archive containing TXT file.

    Returns
    -------
    str
        Contents of a TXT file.
    """
    with zip_file.open(path) as f:
        return f.read().decode("utf-8")


def update_txt(path, text, verbose=True):
    """
    Appends the given text to a specified text file and prints the text.

    Parameters
    ----------
    path : str
        Path to txt file where the text will be appended.
    text : str
        Text to be written to the file.
    verbose : bool, optional
        Indication of whether to printout text. Default is True.
    """
    # Printout text (if applicable)
    if verbose:
        print(text)

    # Update txt file
    with open(path, "a") as file:
        file.write(text + "\n")


# --- Graph Utils ---
def get_leafs(graph):
    """
    Gets all leafs nodes in the given graph.

    Parameters
    ----------
    graph : networkx.Graph
        Graph to be searched.

    Returns
    -------
    List[int]
        Leaf nodes in the given graph.
    """
    return [node for node in graph.nodes if graph.degree[node] == 1]


def search_branching_node(graph, kdtree, root, radius=40):
    """
    Searches for a branching node within distance "radius" from the given
    root node.

    Parameters
    ----------
    graph : networkx.Graph
        Graph to be searched.
    kdtree : scipy.spatial.KDTree
        KDTree containing physical coordinates from a ground truth tracing.
    root : int
        Root of search.
    radius : float, optional
        Distance to search from root. Default is 40.

    Returns
    -------
    root : int
        Root node or closest branching node within distance "radius".
    """
    queue = deque([(root, 0)])
    visited = {root}
    while queue:
        # Visit node
        i, d_i = queue.popleft()
        xyz_i = graph.get_xyz(i)
        if graph.degree[i] > 2:
            dist, _ = kdtree.query(xyz_i)
            if dist < 16:
                return i

        # Update queue
        for j in graph.neighbors(i):
            d_j = d_i + graph.physical_dist(i, j)
            if j not in visited and d_j < radius:
                queue.append((j, d_j))
                visited.add(j)
    return root


# -- GCS Utils --
def is_gcs_path(path):
    """
    Checks if the path is a GCS path.

    Parameters
    ----------
    path : str
        Path to be checked.

    Returns
    -------
    bool
        Indication of whether the path is a GCS path.
    """
    return path.startswith("gs://")


def list_gcs_filenames(bucket_name, prefix, extension=""):
    """
    Lists all files in a GCS bucket with the given extension.

    Parameters
    ----------
    bucket_name : str
        Name of bucket to be searched.
    prefix : str
        Path to location within bucket to be searched.
    extension : str, optional
        File extension of filenames to be listed. Default is an empty string.

    Returns
    -------
    List[str]
        Filenames stored at the GCS path with the given extension.
    """
    bucket = storage.Client().bucket(bucket_name)
    blobs = bucket.list_blobs(prefix=prefix)
    return [blob.name for blob in blobs if extension in blob.name]


def list_gcs_subdirectories(bucket_name, prefix):
    """
    Lists all direct subdirectories of a given prefix in a GCS bucket.

    Parameters
    ----------
    bucket : str
        Name of bucket to be read from.
    prefix : str
        Path to directory in "bucket".

    Returns
    -------
    subdirs : List[str]
         List of direct subdirectories.
    """
    # Load blobs
    storage_client = storage.Client()
    blobs = storage_client.list_blobs(
        bucket_name, prefix=prefix, delimiter="/"
    )
    [blob.name for blob in blobs]

    # Parse directory contents
    prefix_depth = len(prefix.split("/"))
    subdirs = list()
    for prefix in blobs.prefixes:
        is_dir = prefix.endswith("/")
        is_direct_subdir = len(prefix.split("/")) - 1 == prefix_depth
        if is_dir and is_direct_subdir:
            subdirs.append(prefix)
    return subdirs


def read_txt_from_gcs(bucket_name, path):
    """
    Reads a txt file stored in a GCS bucket.

    Parameters
    ----------
    bucket_name : str
        Name of bucket to be read from.
    path : str
        Path to txt file to be read.

    Returns
    -------
    str
        Contents of txt file.
    """
    client = storage.Client()
    bucket = client.bucket(bucket_name)
    blob = bucket.blob(path)
    return blob.download_as_text()


def upload_directory_to_gcs(bucket_name, source_dir, destination_dir):
    """
    Uploads the contents of a local directory to a GCS bucket.

    Parameters
    ----------
    bucket_name : str
        Name of bucket to be read from.
    source_dir : str
        Path to the local directory whose contents should be uploaded.
    destination_dir : str
        Prefix path in the GCS bucket under which the files will be stored.
    """
    client = storage.Client()
    bucket = client.bucket(bucket_name)
    for root, _, files in os.walk(source_dir):
        for filename in files:
            local_path = os.path.join(root, filename)

            # Compute the relative path and GCS destination path
            path = os.path.relpath(local_path, start=source_dir)
            blob_path = os.path.join(destination_dir, path).replace("\\", "/")

            # Upload the file
            blob = bucket.blob(blob_path)
            blob.upload_from_filename(local_path)


# --- S3 Utils ---
def is_s3_path(path):
    """
    Checks if the given path is an S3 path.

    Parameters
    ----------
    path : str
        Path to be checked.

    Returns
    -------
    bool
        Indication of whether the path is an S3 path.
    """
    return path.startswith("s3://")


def list_s3_filenames(bucket_name, prefix, extension=""):
    """
    Lists all object keys in a public S3 bucket under a given prefix,
    optionally filters by file extension.

    Parameters
    ----------
    bucket_name : str
        Name of the S3 bucket.
    prefix : str
        The S3 "directory" prefix to search under.
    extension : str, optional
        File extension to filter by. Default is an empty string, which returns
        all files.

    Returns
    -------
    filenames : List[str]
        List of S3 object keys that match the prefix and extension filter.
    """
    # Create an anonymous client for public buckets
    s3 = boto3.client("s3", config=Config(signature_version=UNSIGNED))
    response = s3.list_objects_v2(Bucket=bucket_name, Prefix=prefix)

    # List all objects under the prefix
    filenames = list()
    if "Contents" in response:
        for obj in response["Contents"]:
            filename = obj["Key"]
            if filename.endswith(extension):
                filenames.append(filename)
    return filenames


def read_txt_from_s3(path):
    """
    Reads a txt file stored in an S3 bucket.

    Parameters
    ----------
    bucket_name : str
        Name of bucket to be read from.
    path : str
        Path to txt file to be read.

    Returns
    -------
    str
        Contents of txt file.
    """
    bucket_name, path = parse_cloud_path(path)
    s3 = boto3.client("s3", config=Config(signature_version=UNSIGNED))
    obj = s3.get_object(Bucket=bucket_name, Key=path)
    return obj["Body"].read().decode("utf-8")


# --- Miscellaneous ---
def compute_weighted_avg(df, column_name):
    """
    Compute the weighted average of a specified column in a DataFrame.

    Parameters
    ----------
    df : pandas.DataFrame
        Input DataFrame containing the target column and a 'SWC Run Length'
        column used as weights.
    column_name : str
        Name of the column for which to compute the weighted average.

    Returns
    -------
    float
        Weighted average of the specified column, ignoring rows where either
        the value or weight is NaN. Returns NaN if the total weight is zero.
    """
    # Extract values
    values = df[column_name]
    weights = df["SWC Run Length"]

    # Ignore NaNs
    mask = values.notna() & weights.notna()
    values = values[mask]
    weights = weights[mask]

    # Compute weighted mean
    if weights.sum() == 0:
        return float("nan")
    else:
        return (values * weights).sum() / weights.sum()


def compute_segmented_run_length(graph, results, name):
    """
    Computes the run length of a graph that was segmented.

    Parameters
    ----------
    graph : LabeledGraph
        Graph to be evaluated.
    results : pandas.DataFrame
        Data frame containing skeleton metrics

    Returns
    -------
    float
        Run length of a graph that was segmented.
    """
    omit_run_length = graph.run_length * results["% Omit Edges"][name] / 100
    split_run_length = graph.run_length * results["% Split Edges"][name] / 100
    return graph.run_length - omit_run_length - split_run_length


def get_segment_id(filename):
    """
    Gets the segment ID correspionding to the given filename, assuming that
    the format of filename is "{segment_id}.{*anything*}.swc".

    Parameters
    ----------
    filename : str
        Name of file to extract segmentation ID from.

    Returns
    -------
    int
        Segment ID.
    """
    return int(filename.split(".")[0])


def load_merged_labels(path):
    """
    Loads segment IDs that are known to contain a merge mistake from the
    given txt file.

    Parameters
    ----------
    path : str
        Path to txt file containing segment IDs.

    Returns
    -------
    List[int]
        Segment IDs that are known to contain a merge mistake.
    """
    df = pd.read_csv(path)
    return list(df["Segment_ID"])


def load_valid_labels(path):
    """
    Loads segment IDs that can be assigned to nodes, accounts for segments
    that may have been removed due to some type of filtering. The default is
    None.

    Parameters
    ----------
    path : str
        Path to txt file containing segment IDs.

    Returns
    -------
    Set[int]
        Segment IDs that can be assigned to nodes.
    """
    valid_labels = set()
    for label_str in read_txt(path).splitlines():
        valid_labels.add(int(label_str.split(".")[0]))
    return valid_labels


def kdtree_query(kdtree, xyz):
    """
    Gets the nearest neighbor of the given xyz coordinate from "kdtree".

    Parameters
    ----------
    kdtree : scipy.spatial.KDTree
        KD-Tree to be searched.
    xyz : ArrayLike
        Coordinate to be queried.

    Returns
    -------
    Tuple[float]
        Coordinate of the nearest neighbor in the given KD-Tree.
    """
    _, idx = kdtree.query(xyz)
    return tuple(kdtree.data[idx])


def parse_cloud_path(path):
    """
    Parses a cloud storage path into its bucket name and key/prefix. Supports
    paths of the form: "{scheme}://bucket_name/prefix" or without a scheme.

    Parameters
    ----------
    path : str
        Path to be parsed.

    Returns
    -------
    bucket_name : str
        Name of the bucket.
    prefix : str
        Cloud prefix.
    """
    # Remove s3:// if present
    if is_s3_path(path):
        path = path[len("s3://"):]

    # Remove gs:// if present
    if is_gcs_path(path):
        path = path[len("gs://"):]

    # Extract bucket and prefix
    parts = path.split("/", 1)
    bucket_name = parts[0]
    prefix = parts[1] if len(parts) > 1 else ""
    return bucket_name, prefix


def sample_once(my_container):
    """
    Samples a single element from "my_container".

    Parameters
    ----------
    my_container : Container
        Container to be sampled from.

    Returns
    -------
    Hashable
        Random element from the given container.
    """
    return sample(my_container, 1)[0]


def to_zipped_point(zip_writer, filename, xyz):
    """
    Writes a point to an SWC file format, which is then stored in a ZIP
    archive.

    Parameters
    ----------
    zip_writer : zipfile.ZipFile
        A ZipFile object that will store the generated SWC file.
    filename : str
        Filename of SWC file.
    xyz : ArrayLike
        Point to be written to SWC file.
    """
    with StringIO() as text_buffer:
        # Preamble
        text_buffer.write("# COLOR 1.0 0.0 0.0")
        text_buffer.write("\n" + "# id, type, z, y, x, r, pid")

        # Write entry
        x, y, z = tuple(xyz)
        text_buffer.write("\n" + f"1 2 {x} {y} {z} 10 -1")

        # Finish
        zip_writer.writestr(filename, text_buffer.getvalue())



"""
Created on Sat May 9 11:00:00 2024

@author: Anna Grim
@email: anna.grim@alleninstitute.org

Code for reading and processing images.

"""

from abc import ABC, abstractmethod
from tifffile import imread

import io
import json
import numpy as np
import os
import tensorstore as ts
import zipfile

from segmentation_skeleton_metrics.utils import util


class Image(ABC):
    """
    Abstract base class for creating image reader subclasses.
    """

    def __init__(self, img_path):
        """
        Instantiates an Image object.

        Parameters
        ----------
        img_path : str
            Path to an image.
        """
        self.img = None
        self.img_path = img_path
        self._load_image()

    @abstractmethod
    def _load_image(self):
        """
        This method should be implemented by subclasses to load the image
        based on the "img_path" attribute.
        """
        pass

    def read(self, voxel, shape):
        """
        Reads a patch from an image given a voxel coordinate and patch shape.

        Parameters
        ----------
        voxel : Tuple[int]
            Voxel coordinate of top-left-front corner of the image patch to be
            read.
        shape : Tuple[int]
            Shape of the image patch to be read.

        Returns
        -------
        numpy.ndarray
            Image patch.
        """
        return self.img[(0, 0, *get_slices(voxel, shape))]

    def read_with_bbox(self, bbox):
        """
        Reads the image patch defined by the given bounding box.

        Parameters
        ----------
        bbox : dict
            Dictionary that contains min and max coordinates of a bounding
            box.

        Returns
        -------
        numpy.ndarray
            Image patch.
        """
        shape = [bbox["max"][i] - bbox["min"][i] for i in range(3)]
        return self.read(bbox["min"], shape)

    def shape(self):
        """
        Gets the shape of image.

        Returns
        -------
        Tuple[int]
            Shape of image.
        """
        return self.img.shape


class TensorStoreImage(Image):
    """
    Class that reads an image with the TensorStore library.
    """

    def __init__(self, img_path, swap_axes=True):
        """
        Instantiates a TensorStore image reader.

        Parameters
        ----------
        img_path : str
            Path to image.
        """
        # Instance attributes
        self.driver = self.get_driver(img_path)
        self.swap_axes = swap_axes

        # Call parent class
        super().__init__(img_path)

    def get_driver(self, img_path):
        """
        Gets the storage driver needed to read the image.

        Returns
        -------
        str
            Storage driver needed to read the image.
        """
        if ".zarr" in img_path:
            return "zarr"
        elif ".n5" in img_path:
            return "n5"
        elif is_precomputed(img_path):
            return "neuroglancer_precomputed"
        else:
            raise ValueError(f"Unsupported image format: {img_path}")

    def _load_image(self):
        """
        Loads image using the TensorStore library.
        """
        # Extract metadata
        bucket_name, path = util.parse_cloud_path(self.img_path)
        storage_driver = get_storage_driver(self.img_path)

        # Load image
        self.img = ts.open(
            {
                "driver": self.driver,
                "kvstore": {
                    "driver": storage_driver,
                    "bucket": bucket_name,
                    "path": path,
                },
                "context": {
                    "cache_pool": {"total_bytes_limit": 1000000000},
                    "cache_pool#remote": {"total_bytes_limit": 1000000000},
                    "data_copy_concurrency": {"limit": 8},
                },
                "recheck_cached_data": "open",
            }
        ).result()

        # Check for Google segmentation
        if "from_google" in self.img_path:
            self.img = self.img[ts.d[:].transpose[3, 2, 1, 0]]

        # Check dimensions
        while self.img.ndim < 5:
            self.img = self.img[ts.newaxis, ...]

    def read(self, voxel, shape):
        """
        Reads a patch from an image given a voxel coordinate and patch shape.

        Parameters
        ----------
        voxel : Tuple[int]
            Voxel coordinate of the top-left-front corner of the image patch
            to be read.
        shape : Tuple[int]
            Shape of the image patch to be read.

        Returns
        -------
        numpy.ndarray
            Image patch.
        """
        return super().read(voxel, shape).read().result()


class TiffImage(Image):
    """
    Class that reads an image with the Tifffile library.
    """

    def __init__(self, img_path, inner_tiff=None, swap_axes=True):
        """
        Instantiates a TiffReader image reader.

        Parameters
        ----------
        img_path : str
            Path to a TIFF image or ZIP archive containing a TIFF image.
        inner_tiff : str or None, optional
            If img_path is a ZIP file, specifies the TIFF filename inside the
            ZIP. Default is None.
        swap_axes : bool, optional
            Indication of whether to swap axes 0 and 2. Default is True.
        """
        # Instance attributes
        self.inner_tiff = inner_tiff
        self.swap_axes = swap_axes

        # Call parent class
        super().__init__(img_path)

    def _load_image(self):
        """
        Loads image using the Tifffile library.
        """
        # Read image
        if self.img_path.lower().endswith(".zip"):
            assert self.inner_tiff is not None, "Must provide TIFF filename!"
            self._load_zipped_image()
        else:
            self.img = imread(self.img_path)

        # Check image dimensions
        while self.img.ndim < 5:
            self.img = self.img[np.newaxis, ...]

        # Swap axes (if applicable)
        if self.swap_axes:
            self.img = np.swapaxes(self.img, 2, 4)

    def _load_zipped_image(self):
        """
        Loads an image in a ZIP archive using the Tifffile library.
        """
        with zipfile.ZipFile(self.img_path, "r") as z:
            # Collect only valid TIFF files, ignoring __MACOSX junk
            tiff_files = [
                f
                for f in z.namelist()
                if f.lower().endswith((".tif", ".tiff"))
                and not os.path.basename(f).startswith("._")
            ]

            # Choose file
            matches = [f for f in tiff_files if f.endswith(self.inner_tiff)]
            if not matches:
                raise FileNotFoundError(f"{self.inner_tiff} not found in ZIP")
            filename = matches[0]

            # Load TIFF
            with z.open(filename) as f:
                self.img = imread(io.BytesIO(f.read()))


# --- Helpers ---
def get_slices(voxel, shape):
    """
    Gets the start and end indices of the chunk to be read.

    Parameters
    ----------
    voxel : Tuple[int]
        Start voxel of image patch to be read.
    shape : Tuple[int]
        Shape of image patch to be read.

    Return
    ------
    Tuple[slice]
        Slice objects used to index into the image.
    """
    return tuple(slice(v, v + d) for v, d in zip(voxel, shape))


def get_storage_driver(img_path):
    """
    Gets the storage driver needed to read the image.

    Parameters
    ----------
    img_path : str
        Image path to be checked.

    Returns
    -------
    str
        Storage driver needed to read the image.
    """
    if util.is_s3_path(img_path):
        return "s3"
    elif util.is_gcs_path(img_path):
        return "gcs"
    else:
        raise ValueError(f"Unsupported path type: {img_path}")


def is_precomputed(img_path):
    """
    Checks if the path points to a Neuroglancer precomputed dataset.

    Parameters
    ----------
    img_path : str
        Path to be checked (can be local, GCS, or S3).

    Returns
    -------
    bool
        True if the path appears to be a Neuroglancer precomputed dataset.
    """
    try:
        # Build kvstore spec
        bucket_name, path = util.parse_cloud_path(img_path)
        kv = {"driver": "gcs", "bucket": bucket_name, "path": path}

        # Open the info file
        store = ts.KvStore.open(kv).result()
        raw = store.read(b"info").result()

        # Only proceed if the key exists and has content
        if raw.state != "missing" and raw.value:
            info = json.loads(raw.value.decode("utf8"))
            is_valid_type = info.get("type") in ("image", "segmentation")
            if isinstance(info, dict) and is_valid_type and "scales" in info:
                return True
        return False
    except Exception:
        return False



"""
Created on Wed Dec 21 19:00:00 2022

@author: Anna Grim
@email: anna.grim@alleninstitute.org

Code that runs of demo of using this library to compute skeleton metrics.

"""

from segmentation_skeleton_metrics.evaluate import evaluate
from segmentation_skeleton_metrics.utils.img_util import TiffImage


def main():
    """
    Evaluates the accuracy of a predicted segmentation by comparing it to a
    set of ground truth skeletons, then reports and saves various performance
    metrics.
    """
    # Initializations
    segmentation = TiffImage(segmentation_path, swap_axes=False)
    evaluate(
        groundtruth_path,
        segmentation,
        output_dir,
        fragments_path=fragments_path,
    )


if __name__ == "__main__":
    # Initializations
    output_dir = "./"
    segmentation_path = "./data/pred_segmentation.tif"
    fragments_path = "./data/pred_swcs.zip"
    groundtruth_path = "./data/target_swcs.zip"

    # Run
    main()



"""
Created on Mon Oct 20 12:00:00 2025

@author: Anna Grim
@email: anna.grim@alleninstitute.org

Implementation of class that computes skeleton-based metrics by comparing a
predicted neuron segmentation to a set of ground truth graphs.

"""

from abc import ABC, abstractmethod
from copy import deepcopy
from collections import deque
from scipy.spatial import KDTree
from tqdm import tqdm

import networkx as nx
import numpy as np
import pandas as pd

from segmentation_skeleton_metrics.utils import util


class SkeletonMetric(ABC):
    """
    Abstract base class for skeleton-based evaluation metrics.
    """

    def __init__(self, verbose=True):
        """
        Instantiates a SkeletonMetric object.

        Parameters
        ----------
        verbose : bool, optional
            Indication of whether to display a progress bar. Default is True.
        """
        # Instance attributes
        self.verbose = verbose

    @abstractmethod
    def __call__(self, gt_graphs):
        """
        Abstract method to be implemented by the subclasses.
        """
        pass

    def get_iterator(self, iterator):
        """
        Gets the iterator wrapped in a progress bar if "verbose" is True,
        otherwise returns the iterator unchanged.

        Parameters
        ----------
        iterator : iterable
            Object to be iterated over.

        Returns
        -------
        iterable
            Object to be iterated over.
        """
        return tqdm(iterator, desc=self.name) if self.verbose else iterator

    def reformat(self, results):
        """
        Converts a dictionary of results into a pandas DataFrame.

        Parameters
        ----------
        results : Dict[str, float]
            Dictionary where keys will become the DataFrame index and values
            are used as the single column data.

        Returns
        -------
        results : pandas.DataFrame
            DataFrame where the indices are the dictionary keys and values are
            stored under a column called "self.name".
        """
        results = pd.DataFrame.from_dict(
            results, orient="index", columns=[self.name]
        )
        return results


# --- Subclasses ---
class SplitEdgePercentMetric(SkeletonMetric):
    """
    A skeleton metric subclass that computes the percentage of split edges.
    """

    def __init__(self, verbose=True):
        """
        Instantiates a SplitEdgePercentMetric object.

        Parameters
        ----------
        verbose : bool, optional
            Indication of whether to display a progress bar. Default is True.
        """
        # Call parent class
        super().__init__(verbose=verbose)

        # Instance attributes
        self.name = "% Split Edges"

    def __call__(self, gt_graphs):
        """
        Computes the percentage of split edges in the given graphs.

        Parameters
        ----------
        gt_graphs : Dict[str, LabeledGraph]
            Graphs to be evaluated.

        Returns
        -------
        results : pandas.DataFrame
            DataFrame where the indices are the dictionary keys and values are
            stored under a column called "self.name".
        """
        results = dict()
        for name, graph in self.get_iterator(gt_graphs.items()):
            num_split_edges = self.count_split_edges(graph)
            results[name] = 100 * num_split_edges / graph.number_of_edges()
        return self.reformat(results)

    @staticmethod
    def count_split_edges(graph):
        """
        Counts the number of split edges in the given graph.

        Parameters
        ----------
        graph : LabeledGraph
            Graph to be evaluated.

        Returns
        -------
        num_split_edges : int
            Number fo split edges in the given graph.
        """
        num_split_edges = 0
        for i, j in nx.dfs_edges(graph):
            is_different = graph.node_labels[i] != graph.node_labels[j]
            is_nonzero = graph.node_labels[i] and graph.node_labels[j]
            if is_different and is_nonzero:
                num_split_edges += 1
        return num_split_edges


class OmitEdgePercentMetric(SkeletonMetric):
    """
    A skeleton metric subclass that computes the percentage of omit edges.
    """

    def __init__(self, verbose=True):
        """
        Instantiates an OmitEdgePercentMetric object.

        Parameters
        ----------
        verbose : bool, optional
            Indication of whether to display a progress bar. Default is True.
        """
        # Call parent class
        super().__init__(verbose=verbose)

        # Instance attributes
        self.name = "% Omit Edges"

    def __call__(self, gt_graphs):
        """
        Computes the percentage of omit edges in the given graphs.

        Parameters
        ----------
        gt_graphs : Dict[str, LabeledGraph]
            Graphs to be evaluated.

        Returns
        -------
        results : pandas.DataFrame
            DataFrame where the indices are the dictionary keys and values are
            stored under a column called "self.name".
        """
        results = dict()
        for name, graph in self.get_iterator(gt_graphs.items()):
            num_omit_edges = self.count_omit_edges(graph)
            omit_edge_percent = 100 * num_omit_edges / graph.number_of_edges()
            results[name] = omit_edge_percent
        return self.reformat(results)

    @staticmethod
    def count_omit_edges(graph):
        """
        Counts the number of omit edges in the given graph.

        Parameters
        ----------
        graph : LabeledGraph
            Graph to be evaluated.

        Returns
        -------
        num_omit_edges : int
            Number fo omit edges in the given graph.
        """
        num_omit_edges = 0
        for i, j in nx.dfs_edges(graph):
            if graph.node_labels[i] == 0 and graph.node_labels[j] == 0:
                num_omit_edges += 1
        return num_omit_edges


class MergedEdgePercentMetric(SkeletonMetric):
    """
    A skeleton metric subclass that computes the percentage of edges that
    are associated with a merge mistake.
    """

    def __init__(self, verbose=True):
        """
        Instantiates a MergedEdgePercentMetric object.

        Parameters
        ----------
        verbose : bool, optional
            Indication of whether to display a progress bar. Default is True.
        """
        # Call parent class
        super().__init__(verbose=verbose)

        # Instance attributes
        self.name = "% Merged Edges"

    def __call__(self, gt_graphs):
        """
        Computes the percentage of merged edges in the given graphs.

        Parameters
        ----------
        gt_graphs : Dict[str, LabeledGraph]
            Graph to be evaluated.

        Returns
        -------
        results : pandas.DataFrame
            DataFrame where the indices are the dictionary keys and values are
            stored under a column called "self.name".
        """
        # Find graphs with common labels
        self.detect_label_intersections(gt_graphs)

        # Compile results
        results = dict()
        for name, graph in gt_graphs.items():
            # Count number of edges associated with a merge
            num_merged_edges = 0
            for label in graph.labels_with_merge:
                num_merged_edges += len(graph.get_nodes_with_label(label)) - 1

            # Compute result
            percent = 100 * num_merged_edges / graph.number_of_edges()
            results[name] = percent
        return self.reformat(results)

    def detect_label_intersections(self, gt_graphs):
        """
        Detects pairs of distinct graphs that contain nodes that share the
        same label.

        Parameters
        ----------
        gt_graphs : Dict[str, LabeledGraph]
            Graphs to be searched for intersecting labels.
        """
        visited = set()
        for name1, graph1 in self.get_iterator(gt_graphs.items()):
            # Search other graphs for label intersections
            for name2, graph2 in gt_graphs.items():
                names = frozenset((name1, name2))
                if name1 != name2 and names not in visited:
                    visited.add(names)
                    labels1 = set(graph1.get_node_labels())
                    labels2 = set(graph2.get_node_labels())
                    for label in labels1.intersection(labels2):
                        # Check if intersection is meaningful
                        num_nodes1 = len(graph1.get_nodes_with_label(label))
                        num_nodes2 = len(graph2.get_nodes_with_label(label))
                        if num_nodes1 > 50 and num_nodes2 > 50:
                            graph1.labels_with_merge.add(label)
                            graph2.labels_with_merge.add(label)


class SplitCountMetric(SkeletonMetric):
    """
    A skeleton metric subclass that counts the number of splits.
    """

    def __init__(self, verbose=True):
        """
        Instantiates a SplitCountMetric object.

        Parameters
        ----------
        verbose : bool, optional
            Indication of whether to display a progress bar. Default is True.
        """
        # Call parent class
        super().__init__(verbose=verbose)

        # Instance attributes
        self.name = "# Splits"

    def __call__(self, gt_graphs):
        """
        Counts the number of split mistakes in each of the given graphs.

        Parameters
        ----------
        gt_graphs : Dict[str, LabeledGraph]
            Graphs to be evaluated.

        Results
        -------
        results : pandas.DataFrame
            DataFrame where the indices are the dictionary keys and values are
            stored under a column called "self.name".
        """
        results = dict()
        for name, graph in self.get_iterator(gt_graphs.items()):
            num_splits = max(len(graph.get_node_labels()) - 1, 0)
            results[name] = int(num_splits)
        return self.reformat(results)


class MergeCountMetric(SkeletonMetric):
    """
    A skeleton metric subclass that counts the number merges.
    """

    merge_dist_threshold = 50

    def __init__(self, verbose=True):
        """
        Instantiates a MergeCountMetric object.

        Parameters
        ----------
        verbose : bool, optional
            Indication of whether to display a progress bar. Default is True.
        """
        # Call parent class
        super().__init__(verbose=verbose)

        # Instance attributes
        self.fragments_with_merge = set()
        self.merge_sites = list()
        self.name = "# Merges"

    # --- Core Routines ---
    def __call__(self, gt_graphs, fragment_graphs):
        """
        Counts the number of split merges in each of the given ground truth
        graphs.

        Parameters
        ----------
        gt_graphs : Dict[str, LabeledGraph]
            Graphs to be evaluated.
        fragment_graphs : Dict[str, FragmentGraph]
            Graphs corresponding to the predicted segmentation.

        Results
        -------
        results : pandas.DataFrame
            DataFrame where the indices are the dictionary keys and values are
            stored under a column called "self.name".
        """
        # Main
        for gt_graph in self.iterator(gt_graphs.values()):
            # Build ground truth kd-tree
            gt_graph.init_kdtree()

            # Search intersecting fragments
            labels = gt_graph.get_node_labels()
            for fragment_graph in fragment_graphs.values():
                if fragment_graph.label in labels:
                    self.search_for_merges(gt_graph, fragment_graph)

        # Postprocess merge sites
        self.remove_repeat_merge_sites()

        # Compile results
        results = dict()
        for name in gt_graphs:
            if len(self.merge_sites) > 0:
                num_merges = (self.merge_sites["GroundTruth_ID"] == name).sum()
            else:
                num_merges = 0
            results[name] = num_merges
        return self.reformat(results)

    def search_for_merges(self, gt_graph, fragment_graph):
        """
        Searches for potential merge errors in a fragment graph by comparing
        it to a ground truth graph.

        Parameters
        ----------
        gt_graph : LabeledGraph
            Graph to be evaluated.
        fragment_graph : FragmentGraph
            Graph corresponding to a segment in the predicted segmentation.
        """
        visited = set()
        for leaf in util.get_leafs(fragment_graph):
            # Check whether to visit
            if leaf in visited or visited.add(leaf):
                continue

            # Find closet node in ground truth
            xyz = fragment_graph.get_xyz(leaf)
            dist, _ = gt_graph.kdtree.query(xyz)

            # Check if distance to ground truth flags a merge mistake
            if dist > MergeCountMetric.merge_dist_threshold:
                self.find_merge_site(gt_graph, fragment_graph, leaf, visited)

    def find_merge_site(self, gt_graph, fragment_graph, source, visited):
        """
        Traverses fragment graph from a source node to locate and verify
        potential merge sites relative to the ground truth graph.

        Parameters
        ----------
        gt_graph : LabeledGraph
            Graphs to be evaluated.
        fragment_graphs : FragmentGraph
            Graph corresponding to a segment in the predicted segmentation.
        source : int
            Starting node ID in the fragment graph from which to begin
            traversal.
        visited : Set[int]
            Node IDs from "fragment_graphs" that have already been visited,
            used to avoid redundant exploration.
        """
        queue = deque([source])
        visited.add(source)
        while len(queue) > 0:
            # Visit node
            i = queue.pop()
            xyz_i = fragment_graph.get_xyz(i)
            dist_i, gt_node = gt_graph.kdtree.query(xyz_i)
            if dist_i < 6:
                self.verify_site(gt_graph, fragment_graph, gt_node, i)
                break

            # Update queue
            for j in fragment_graph.neighbors(i):
                if j not in visited:
                    queue.append(j)
                    visited.add(j)

    def verify_site(self, gt_graph, fragment_graph, gt_node, fragment_node):
        """
        Verifies whether a given site in a fragment graph corresponds to a
        merge mistake relative to the ground truth graph. If so, the site is
        saved in an internal data structure.

        Parameters
        ----------
        gt_graph : LabeledGraph
            Graph to be evaluated.
        fragment_graph : FragmentGraph
            Graph corresponding to a segment in the predicted segmentation.
        gt_node : int
            Node ID in the ground truth graph corresponding to the site.
        fragment_node : int
            Node ID in the fragment graph corresponding to the candidate site.
        """
        # Check if pass through site without merge mistake
        if self.is_nonmerge_pass_thru(gt_graph, fragment_graph, gt_node):
            return

        # Move site to nearby branching point if possible
        fragment_node = util.search_branching_node(
            fragment_graph, gt_graph.kdtree, fragment_node
        )

        # Record site as merge mistake
        voxel = fragment_graph.voxels[fragment_node]
        xyz = fragment_graph.get_xyz(fragment_node)

        self.fragments_with_merge.add(fragment_graph.name)
        self.merge_sites.append(
            {
                "Fragment_Name": fragment_graph.name,
                "Segment_ID": fragment_graph.segment_id,
                "GroundTruth_ID": gt_graph.name,
                "Voxel": tuple(map(int, voxel)),
                "World": tuple([float(round(t, 2)) for t in xyz]),
                "Added Cable Length (μm)": 0.0,
            }
        )

    def is_nonmerge_pass_thru(self, gt_graph, fragment_graph, gt_node):
        """
        Determines whether a ground truth node belongs to a small connected
        component of the same label in the ground truth graph, indicating a
        likely non-merge pass-through.

        Parameters
        ----------
        gt_graph : LabeledGraph
            Graph to be evaluated.
        fragment_graph : FragmentGraph
            Graph corresponding to a segment in the predicted segmentation.
        gt_node : int
            Node ID in the ground truth graph to evaluate.

        Returns
        -------
        bool
            True is the node is a likely non-merge detection; otherwise, the
            site is considered to be a merge mistake.
        """
        nodes = gt_graph.get_nodes_with_label(fragment_graph.label)
        subgraph = gt_graph.subgraph(nodes)
        for nodes_cc in nx.connected_components(subgraph):
            if gt_node in nodes_cc:
                return len(nodes_cc) < 50
        return True

    # --- Helpers ---
    def add_merge_site_names(self):
        """
        Assigns unique name to each detected merge site.
        """
        row_names = list()
        for i, _ in enumerate(self.merge_sites.index, 1):
            row_names.append(f"merge-{i}.swc")
        self.merge_sites.index = row_names
        self.merge_sites.index.name = "Merge_ID"

    def remove_repeat_merge_sites(self):
        """
        Removes spatially redundant merge sites within a fixed distance
        threshold.
        """
        if len(self.merge_sites) > 0:
            # Build kdtree from merge sites
            kdtree = KDTree([s["World"] for s in self.merge_sites])

            # Search for repeat sites
            rm_idxs = set()
            for i, site in enumerate(self.merge_sites):
                if i not in rm_idxs:
                    idxs = kdtree.query_ball_point(site["World"], 40)
                    idxs.remove(i)
                    rm_idxs |= set(idxs)

            # Remove repeat sites
            self.merge_sites = pd.DataFrame(self.merge_sites).drop(rm_idxs)
            self.add_merge_site_names()
        else:
            self.merge_sites = pd.DataFrame()


class ERLMetric(SkeletonMetric):
    """
    A skeleton metric subclass that computes the expected run length (ERL).
    """

    def __init__(self, verbose):
        """
        Instantiates an ERL object.

        Parameters
        ----------
        verbose : bool, optional
            Indication of whether to display a progress bar. Default is True.
        """
        # Call parent class
        super().__init__(verbose=verbose)

        # Instance attributes
        self.name = "ERL"

    def __call__(self, gt_graphs):
        """
        Comptues the expected run length (ERL) of the given graphs.

        gt_graphs : Dict[str, LabeledGraph]
            Graphs to be evaluated.

        Returns
        -------
        results : pandas.DataFrame
            DataFrame where the indices are the dictionary keys and values are
            stored under a column called "self.name".
        """
        results = dict()
        for name, graph in self.get_iterator(gt_graphs.items()):
            results[name] = round(self.compute_graph_erl(graph), 2)
        return self.reformat(results)

    @staticmethod
    def compute_graph_erl(graph):
        """
        Computes the ERL of the given graph.

        Parameters
        ----------
        graph : LabeledGraph
            Graph to be evaluated.

        Returns
        -------
        float
            ERL of the given graph.
        """
        wgts = list()
        run_lengths = list()
        for label in graph.get_node_labels():
            # Compute run length for label
            nodes = graph.get_nodes_with_label(label)
            run_length = graph.run_length_from(nodes[0])
            graph.labeled_run_length += run_length

            # Update
            wgts.append(run_length)
            run_lengths.append(
                0 if label in graph.labels_with_merge else run_length
            )
        return np.average(run_lengths, weights=wgts) if len(wgts) > 0 else 0


# --- Derived Skeleton Metrics ---
class SplitRateMetric(SkeletonMetric):
    """
    A skeleton metric subclass that computes split rate as µm / num_splits.
    """

    def __init__(self, verbose=True):
        """
        Instantiates a SplitRateMetric object.

        Parameters
        ----------
        verbose : bool, optional
            Indication of whether to display a progress bar. Default is True.
        """
        # Call parent class
        super().__init__(verbose=verbose)

        # Instance attributes
        self.name = "Split Rate"

    def __call__(self, gt_graphs, results):
        """
        Computes split rates for the given graphs.

        Parameters
        ----------
        gt_graphs : Dict[str, LabeledGraph]
            Graphs to be evaluated.
        results : pandas.DataFrame
            Data frame containing the skeleton metric results computed so far.

        Returns
        -------
        results : pandas.DataFrame
            DataFrame where the indices are the dictionary keys and values are
            stored under a column called "self.name".
        """
        new_results = dict()
        for name, graph in self.get_iterator(gt_graphs.items()):
            if results["# Splits"][name] > 0:
                rl = util.compute_segmented_run_length(graph, results, name)
                new_results[name] = round(rl / results["# Splits"][name], 2)
            else:
                new_results[name] = np.nan
        return self.reformat(new_results)


class MergeRateMetric(SkeletonMetric):
    """
    A skeleton metric subclass that computes merge rate as µm / num_merges.
    """

    def __init__(self, verbose=True):
        """
        Instantiates a MergeRateMetric object.

        Parameters
        ----------
        verbose : bool, optional
            Indication of whether to display a progress bar. Default is True.
        """
        # Call parent class
        super().__init__(verbose=verbose)

        # Instance attributes
        self.name = "Merge Rate"

    def __call__(self, gt_graphs, results):
        """
        Computes merge rates for the given graphs.

        Parameters
        ----------
        gt_graphs : Dict[str, LabeledGraph]
            Graphs to be evaluated.
        results : pandas.DataFrame
            Data frame containing the skeleton metric results computed so far.

        Returns
        -------
        results : pandas.DataFrame
            DataFrame where the indices are the dictionary keys and values are
            stored under a column called "self.name".
        """
        new_results = dict()
        for name, graph in self.get_iterator(gt_graphs.items()):
            if results["# Merges"][name] > 0:
                rl = util.compute_segmented_run_length(graph, results, name)
                new_results[name] = round(rl / results["# Merges"][name], 2)
            else:
                new_results[name] = np.nan
        return self.reformat(new_results)


class EdgeAccuracyMetric(SkeletonMetric):
    """
    A skeleton metric subclass that computes edge accuracy.
    """

    def __init__(self, verbose=True):
        """
        Instantiates an EdgeAccuracyMetric object.

        Parameters
        ----------
        verbose : bool, optional
            Indication of whether to display a progress bar. Default is True.
        """
        # Call parent class
        super().__init__(verbose=verbose)

        # Instance attributes
        self.name = "Edge Accuracy"

    def __call__(self, gt_graphs, results):
        """
        Computes the edge accuracy of the given graphs.

        Parameters
        ----------
        gt_graphs : Dict[str, LabeledGraph]
            Graphs to be evaluated.
        results : pandas.DataFrame
            Data frame containing the skeleton metric results computed so far.

        Returns
        -------
        results : pandas.DataFrame
            DataFrame where the indices are dictionary keys and values are
            stored under a column called "self.name".
        """
        new_results = dict()
        for idx in self.get_iterator(results.index):
            edge_accuracy = 100 - (
                results["% Split Edges"].loc[idx]
                + results["% Omit Edges"].loc[idx]
                + results["% Merged Edges"].loc[idx]
            )
            new_results[idx] = round(edge_accuracy, 2)
        return self.reformat(new_results)


class NormalizedERLMetric(SkeletonMetric):
    """
    A skeleton metric subclass that computes normalized expected run
    length (ERL).
    """

    def __init__(self, verbose=True):
        """
        Instantiates a NormalizedERLMetric object.

        Parameters
        ----------
        verbose : bool, optional
            Indication of whether to display a progress bar. Default is True.
        """
        # Call parent class
        super().__init__(verbose=verbose)

        # Instance attributes
        self.name = "Normalized ERL"

    def __call__(self, gt_graphs, results):
        """
        Computes the normalized ERL of the given graphs.

        Parameters
        ----------
        gt_graphs : Dict[str, LabeledGraph]
            Graphs to be evaluated.
        results : pandas.DataFrame
            Data frame containing the skeleton metric results computed so far.

        Returns
        -------
        results : pandas.DataFrame
            DataFrame where the indices are the dictionary keys and values are
            stored under a column called "self.name".
        """
        new_results = dict()
        for name, graph in self.get_iterator(gt_graphs.items()):
            normalized_erl = results["ERL"][name] / graph.run_length
            new_results[name] = round(normalized_erl, 4)
        return self.reformat(new_results)


class AddedCableLengthMetric(SkeletonMetric):
    """
    A skeleton metric subclass that computes added cable length.
    """

    def __init__(self, verbose=True):
        """
        Instantiates an AddedCableLengthMetric object.

        Parameters
        ----------
        verbose : bool, optional
            Indication of whether to display a progress bar. Default is True.
        """
        # Call parent class
        super().__init__(verbose=verbose)

        # Instance attributes
        self.name = "Added Cable Length (μm)"

    def __call__(self, gt_graphs, fragment_graphs, merge_sites):
        """
        Computes the normalized ERL of the given graphs.

        Parameters
        ----------
        gt_graphs : Dict[str, LabeledGraph]
            Graphs to be evaluated.
        fragment_graphs : Dict[str, FragmentGraph]
            Graphs corresponding to the predicted segmentation.
        merge_sites : pandas.DataFrame
            Data frame containing detected merge sites.

        Returns
        -------
        results : pandas.DataFrame
            DataFrame where the indices are the dictionary keys and values are
            stored under a column called "self.name".
        """
        # Check if merge sites is non-empty
        if len(merge_sites) == 0:
            return None

        # Compute metric
        pair_to_length = dict()
        for i in self.get_iterator(merge_sites.index):
            # Extract site info
            fragment_name = str(merge_sites["Fragment_Name"][i])
            gt_id = merge_sites["GroundTruth_ID"][i]
            pair_id = (fragment_name, gt_id)

            # Check wheter to visit
            if pair_id in pair_to_length:
                merge_sites.loc[i, self.name] = pair_to_length[pair_id]
            else:
                # Get graphs
                gt_graph = gt_graphs[gt_id]
                fragment_graph = deepcopy(fragment_graphs[fragment_name])

                # Compute metric
                pair_to_length[pair_id] = self.compute_added_length(
                    gt_graph, fragment_graph
                )
                merge_sites.loc[i, self.name] = pair_to_length[pair_id]

    def compute_added_length(self, gt_graph, fragment_graph):
        """
        Computes the total cable length of fragment components that are not
        sufficiently close to the ground-truth graph.

        Parameters
        ----------
        gt_graph : LabeledGraph
            Graph containing merge mistake.
        fragment_graph : FragmentGraph
            Fragment that is merged to the given ground truth graph.

        Returns
        -------
        cable_length : float
            Total cable length of fragment components that remain after pruning
            nodes near the ground-truth graph.
        """
        # Remove nodes close to ground truth
        xyz_arr = fragment_graph.voxels * fragment_graph.anisotropy
        dists, _ = gt_graph.kdtree.query(xyz_arr)
        max_dist = MergeCountMetric.merge_dist_threshold
        fragment_graph.remove_nodes_from(np.where(dists < max_dist)[0])

        # Compute cable length
        cable_length = 0
        for nodes in map(list, nx.connected_components(fragment_graph)):
            cable_length += fragment_graph.run_length_from(nodes[0])
        return round(float(cable_length), 2)
